import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Tuple, TypeVar, Union

from golem.exceptions import GolemException
from golem.resources import (
    Activity,
    Agreement,
    Allocation,
    DemandData,
    Proposal,
    ProposalData,
    Script,
)
from golem.resources.activity import commands

logger = logging.getLogger(__name__)


class Batch:
    def __init__(self, activity) -> None:
        self._script = Script()
        self._activity = activity

    def deploy(self, deploy_args: Optional[commands.ArgsDict] = None):
        self._script.add_command(commands.Deploy(deploy_args))

    def start(self):
        self._script.add_command(commands.Start())

    def send_file(self, src_path: str, dst_path: str):
        self._script.add_command(commands.SendFile(src_path, dst_path))

    def download_file(self, src_path: str, dst_path: str):
        self._script.add_command(commands.DownloadFile(src_path, dst_path))

    def run(
        self,
        command: Union[str, List[str]],
        *,
        shell: Optional[bool] = None,
        shell_cmd: str = "/bin/sh",
    ):
        self._script.add_command(commands.Run(command, shell=shell, shell_cmd=shell_cmd))

    async def __call__(self):
        pooling_batch = await self._activity.execute_script(self._script)
        return await pooling_batch.wait()


class WorkContext:
    def __init__(self, activity: Activity):
        self._activity = activity

    async def deploy(
        self, deploy_args: Optional[commands.ArgsDict] = None, timeout: Optional[timedelta] = None
    ):
        pooling_batch = await self._activity.execute_commands(commands.Deploy(deploy_args))
        await pooling_batch.wait(timeout)

    async def start(self):
        pooling_batch = await self._activity.execute_commands(commands.Start())
        await pooling_batch.wait()

    async def terminate(self):
        await self._activity.destroy()

    async def run(
        self,
        command: Union[str, List[str]],
        *,
        shell: Optional[bool] = None,
        shell_cmd: str = "/bin/sh",
    ):
        return await self._activity.execute_commands(
            commands.Run(command, shell=shell, shell_cmd=shell_cmd)
        )

    async def create_batch(self) -> Batch:
        return Batch(self._activity)

    @property
    def activity(self) -> Activity:
        return self._activity

    async def get_provider_id(self):
        """Get the node id of the provider running this context."""
        return await self.activity.agreement.proposal.get_provider_id()

    async def get_provider_name(self):
        """Get the node name of the provider running this context."""
        return await self.activity.agreement.proposal.get_provider_name()


@dataclass
class WorkResult:
    result: Optional[Any] = None
    exception: Optional[Exception] = None
    extras: Dict = field(default_factory=dict)


WORK_PLUGIN_FIELD_NAME = "_work_plugins"

Work = Callable[[WorkContext], Awaitable[Optional[WorkResult]]]

DoWorkCallable = Callable[[Work], Awaitable[WorkResult]]


class ManagerEvent(ABC):
    pass


class ManagerException(GolemException):
    pass


class ManagerPluginException(ManagerException):
    pass


class Manager(ABC):
    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass


TPlugin = TypeVar("TPlugin")


class NetworkManager(Manager, ABC):
    pass


class PaymentManager(Manager, ABC):
    @abstractmethod
    async def get_allocation(self) -> Allocation:
        ...


class DemandManager(Manager, ABC):
    @abstractmethod
    async def get_initial_proposal(self) -> Proposal:
        ...


class ProposalManager(Manager):
    @abstractmethod
    async def get_draft_proposal(self) -> Proposal:
        ...


class AgreementManager(Manager, ABC):
    @abstractmethod
    async def get_agreement(self) -> Agreement:
        ...


class ActivityManager(Manager, ABC):
    @abstractmethod
    async def get_activity(self) -> Activity:
        ...


class WorkManager(Manager, ABC):
    pass


class RejectProposal(ManagerPluginException):
    pass


class ProposalNegotiator(ABC):
    @abstractmethod
    def __call__(
        self, demand_data: DemandData, proposal_data: ProposalData
    ) -> Union[Awaitable[Optional[RejectProposal]], Optional[RejectProposal]]:
        ...


class ProposalManagerPlugin(ABC):
    _get_proposal: Callable[[], Awaitable[Proposal]]

    def set_proposal_callback(self, get_proposal: Callable[[], Awaitable[Proposal]]):
        self._get_proposal = get_proposal

    @abstractmethod
    async def get_proposal(self) -> Proposal:
        ...

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass


ProposalScoringResult = Sequence[Optional[float]]


class ProposalScorer(ABC):
    @abstractmethod
    def __call__(
        self, proposals_data: Sequence[ProposalData]
    ) -> Union[Awaitable[ProposalScoringResult], ProposalScoringResult]:
        ...


ScorerWithOptionalWeight = Union[ProposalScorer, Tuple[float, ProposalScorer]]


class WorkManagerPlugin(ABC):
    @abstractmethod
    def __call__(self, do_work: DoWorkCallable) -> DoWorkCallable:
        ...


PricingCallable = Callable[[ProposalData], Optional[float]]



import asyncio
import logging
from datetime import timedelta
from typing import List, Optional

from golem.managers.proposal.plugins.buffer import Buffer
from golem.managers.proposal.plugins.scoring.mixins import ProposalScoringMixin
from golem.resources import Proposal
from golem.utils.asyncio import create_task_with_logging
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


class ScoringBuffer(ProposalScoringMixin, Buffer):
    def __init__(
        self,
        update_interval: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._scored: List[Proposal] = []
        self._scored_condition = asyncio.Condition()
        self._background_loop_task: Optional[asyncio.Task] = None

        self._items_requested_event = asyncio.Event()
        self._update_interval = update_interval

    @trace_span()
    async def start(self) -> None:
        await super().start()

        self._background_loop_task = create_task_with_logging(
            self._background_loop(), trace_id=get_trace_id_name(self, "background-loop")
        )

    @trace_span()
    async def stop(self) -> None:
        await super().stop()

        if self._background_loop_task is not None:
            self._background_loop_task.cancel()
            self._background_loop_task = None

    def is_started(self) -> bool:
        is_started = super().is_started()
        return (
            is_started
            and self._background_loop_task is not None
            and not self._background_loop_task.done()
        )

    async def _background_loop(self):
        # TODO: Rework this flow to:
        #  - Use composition instead of inheritance
        #  - Get rid of loop checking state for proper state
        #     - composable Buffer should expose async context that exposes its items
        #       and items asyncio.Condition

        while True:
            logger.debug("Waiting for any requested items...")
            await self._items_requested_event.wait()
            logger.debug("Waiting for any requested items done")

            keep_retrying = True
            while keep_retrying:  # FIXME: Needs refactor too
                logger.debug("Waiting up to %s for all requested items...", self._update_interval)
                try:
                    await asyncio.wait_for(
                        self._requests_queue.join(),
                        timeout=self._update_interval.total_seconds(),
                    )
                except asyncio.TimeoutError:
                    logger.debug(
                        "Waiting up to %s for all requested items failed with timeout, trying to"
                        "update anyways...",
                        self._update_interval,
                    )
                else:
                    logger.debug(
                        "Waiting up to %s for all requested items done", self._update_interval
                    )
                    keep_retrying = False
                    self._items_requested_event.clear()

                async with self._buffered_condition:
                    if not self._buffered:
                        logger.debug("Update not needed, as no items were buffered in the meantime")
                        continue

                    items_to_score = self._buffered[:]
                    self._buffered.clear()

                async with self._scored_condition:
                    scored_proposals = await self.do_scoring(self._scored + items_to_score)
                    self._scored = [proposal for _, proposal in scored_proposals]

                    logger.debug("Item collection updated %s", scored_proposals)

                    self._scored_condition.notify_all()

    async def _get_item(self):
        async with self._scored_condition:
            if self._get_items_count() == 0:  # This supports lazy (not at start) buffer filling
                logger.debug("No items to get, requesting fill")
                self._handle_item_requests()

            logger.debug("Waiting for any item to pick...")

            await self._scored_condition.wait_for(lambda: 0 < len(self._scored))
            item = self._scored.pop()

            # Check if we need to request any additional items
            if self._get_items_count() < self._min_size:
                self._handle_item_requests()

            return item

    def _handle_item_requests(self) -> None:
        super()._handle_item_requests()

        self._items_requested_event.set()

    def _get_items_count(self):
        items_count = super()._get_items_count()

        return items_count + len(self._scored)



import logging
from datetime import timedelta
from typing import Optional

from golem.payload import defaults
from golem.resources import LinearCoeffs, ProposalData

logger = logging.getLogger(__name__)


class LinearAverageCostPricing:
    def __init__(self, average_cpu_load: float, average_duration: timedelta) -> None:
        self._average_cpu_load = average_cpu_load
        self._average_duration = average_duration

    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        coeffs = LinearCoeffs.from_proposal_data(proposal_data)

        if coeffs is None:
            return None

        return self._calculate_cost(coeffs)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"average_cpu_load={self._average_cpu_load}, "
            f"average_duration={self._average_duration})"
        )

    def _calculate_cost(self, coeffs: LinearCoeffs) -> float:
        average_duration_sec = self._average_duration.total_seconds()

        average_duration_cost = coeffs.price_duration_sec * average_duration_sec
        average_cpu_cost = coeffs.price_cpu_sec * self._average_cpu_load * average_duration_sec
        average_initial_price = coeffs.price_initial / average_duration_sec

        return average_duration_cost + average_cpu_cost + average_initial_price


class LinearPerCpuAverageCostPricing(LinearAverageCostPricing):
    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        coeffs = LinearCoeffs.from_proposal_data(proposal_data)
        cpu_count = proposal_data.properties.get(defaults.INF_CPU_THREADS)

        if coeffs is None or not cpu_count:
            return None

        coeffs.price_duration_sec /= cpu_count
        coeffs.price_initial /= cpu_count

        return self._calculate_cost(coeffs)



import asyncio
import logging
from decimal import Decimal
from typing import List, Optional

from golem.managers.base import PaymentManager
from golem.node import GolemNode
from golem.payload.defaults import DEFAULT_PAYMENT_DRIVER, DEFAULT_PAYMENT_NETWORK
from golem.resources import (
    AgreementClosed,
    Allocation,
    DebitNote,
    Invoice,
    NewAgreement,
    NewDebitNote,
    NewInvoice,
)
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class PayAllPaymentManager(PaymentManager):
    def __init__(
        self,
        golem: GolemNode,
        budget: float,
        network: str = DEFAULT_PAYMENT_NETWORK,
        driver: str = DEFAULT_PAYMENT_DRIVER,
    ):
        self._golem = golem
        self._budget = budget
        self._network = network
        self._driver = driver

        self._allocation: Optional[Allocation] = None

        self._opened_agreements_count = 0
        self._closed_agreements_count = 0
        self._payed_invoices_count = 0

        self._event_handlers: List = []

    @trace_span("Starting PayAllPaymentManager", log_level=logging.INFO)
    async def start(self):
        # TODO: Add stop with event_bus.off()
        self._event_handlers.extend(
            [
                await self._golem.event_bus.on(NewInvoice, self._pay_invoice_if_received),
                await self._golem.event_bus.on(NewDebitNote, self._pay_debit_note_if_received),
                await self._golem.event_bus.on(NewAgreement, self._increment_opened_agreements),
                await self._golem.event_bus.on(AgreementClosed, self._increment_closed_agreements),
            ]
        )

    @trace_span("Stopping PayAllPaymentManager", log_level=logging.INFO)
    async def stop(self):
        try:
            await self.wait_for_invoices()
        except RuntimeError:
            pass

        for event_handler in self._event_handlers:
            await self._golem.event_bus.off(event_handler)

    @trace_span()
    async def _create_allocation(self) -> None:
        self._allocation = await Allocation.create_any_account(
            self._golem, Decimal(self._budget), self._network, self._driver
        )

        # TODO: We should not rely on golem node with cleanups, manager should do it by itself
        self._golem.add_autoclose_resource(self._allocation)

    @trace_span("Getting allocation", show_results=True, log_level=logging.INFO)
    async def get_allocation(self) -> "Allocation":
        # TODO handle NoMatchingAccount
        if self._allocation is None:
            await self._create_allocation()

        # TODO fix type
        return self._allocation  # type: ignore[return-value]

    @trace_span("Waiting for invoices", log_level=logging.INFO)
    async def wait_for_invoices(self):
        for _ in range(60):
            if (
                self._opened_agreements_count
                == self._closed_agreements_count
                == self._payed_invoices_count
            ):
                logger.info("All invoices paid")
                return

            await asyncio.sleep(1)

        # TODO: Add list of agreements without payment
        raise RuntimeError("Waiting for invoices failed with timeout!")

    async def _increment_opened_agreements(self, event: NewAgreement):
        self._opened_agreements_count += 1

    async def _increment_closed_agreements(self, event: AgreementClosed):
        self._closed_agreements_count += 1

    @trace_span(show_arguments=True)
    async def _accept_invoice(self, invoice: Invoice) -> None:
        assert self._allocation is not None  # TODO think of a better way
        await invoice.accept_full(self._allocation)
        await invoice.get_data(force=True)
        self._payed_invoices_count += 1

        logger.info(f"Invoice `{invoice.id}` for agreement `{invoice.data.agreement_id}` accepted")

    @trace_span(show_arguments=True)
    async def _accept_debit_note(self, debit_note: DebitNote) -> None:
        assert self._allocation is not None  # TODO think of a better way
        await debit_note.accept_full(self._allocation)
        await debit_note.get_data(force=True)

        logger.info(
            f"DebitNote `{debit_note.id}` for agreement `{debit_note.data.agreement_id}` accepted"
        )

    @trace_span()
    async def _pay_invoice_if_received(self, event: NewInvoice) -> None:
        invoice = event.resource

        if (await invoice.get_data(force=True)).status == "RECEIVED":
            logger.debug("Accepting invoice `%s`: %s", invoice, await invoice.get_data())
            await self._accept_invoice(invoice)

    @trace_span()
    async def _pay_debit_note_if_received(self, event: NewDebitNote) -> None:
        debit_note = event.resource

        if (await debit_note.get_data(force=True)).status == "RECEIVED":
            logger.debug("Accepting debit note `%s`: %s", debit_note, await debit_note.get_data())
            await self._accept_debit_note(debit_note)



import inspect
from datetime import datetime
from typing import List, Optional, Sequence, Tuple, cast

from golem.managers.base import ScorerWithOptionalWeight
from golem.payload import PayloadSyntaxParser, Properties
from golem.resources import Proposal, ProposalData
from golem.utils.logging import trace_span


class ProposalScoringMixin:
    def __init__(
        self,
        proposal_scorers: Optional[Sequence[ScorerWithOptionalWeight]] = None,
        *args,
        **kwargs,
    ) -> None:
        self._demand_offer_parser = PayloadSyntaxParser.get_instance()
        self._proposal_scorers: List[ScorerWithOptionalWeight] = (
            list(proposal_scorers) if proposal_scorers is not None else []
        )

        super().__init__(*args, **kwargs)

    async def do_scoring(self, proposals: Sequence[Proposal]) -> List[Tuple[float, Proposal]]:
        proposals_data = await self._get_proposals_data_from_proposals(proposals)
        proposal_scores = await self._run_scorers(proposals_data)

        scored_proposals = self._calculate_proposal_score(proposals, proposal_scores)
        scored_proposals.sort(key=lambda x: x[0], reverse=True)

        return scored_proposals

    @trace_span()
    async def _run_scorers(
        self, proposals_data: Sequence[ProposalData]
    ) -> Sequence[Tuple[float, Sequence[float]]]:
        proposal_scores: List[Tuple[float, Sequence[float]]] = []

        for scorer in self._proposal_scorers:
            if isinstance(scorer, (list, tuple)):
                weight, scorer = scorer
            else:
                weight = 1

            scorer_scores = scorer(proposals_data)

            if inspect.isawaitable(scorer_scores):
                scorer_scores = await scorer_scores

            proposal_scores.append((weight, scorer_scores))  # type: ignore[arg-type]

        return proposal_scores

    def _calculate_proposal_score(
        self,
        proposals: Sequence[Proposal],
        scorer_scores: Sequence[Tuple[float, Sequence[float]]],
    ) -> List[Tuple[float, Proposal]]:
        # FIXME: can this be refactored?
        return [
            (
                self._calculate_weighted_score(
                    self._transpose_scorer_scores(proposal_index, scorer_scores)
                ),
                proposal,
            )
            for proposal_index, proposal in enumerate(proposals)
        ]

    def _calculate_weighted_score(
        self, proposal_weighted_scores: Sequence[Tuple[float, float]]
    ) -> float:
        if not proposal_weighted_scores:
            return 0

        weighted_sum = sum(pws[0] * pws[1] for pws in proposal_weighted_scores)
        weights_sum = sum(pws[0] for pws in proposal_weighted_scores)

        return weighted_sum / weights_sum

    def _transpose_scorer_scores(
        self, proposal_index: int, scorer_scores: Sequence[Tuple[float, Sequence[float]]]
    ) -> Sequence[Tuple[float, float]]:
        # FIXME: can this be refactored?
        return [
            (scorer_weight, scorer_scores[proposal_index])
            for scorer_weight, scorer_scores in scorer_scores
            if scorer_scores[proposal_index] is not None
        ]

    # FIXME: This should be already provided by low level
    async def _get_proposals_data_from_proposals(
        self, proposals: Sequence[Proposal]
    ) -> Sequence[ProposalData]:
        result = []

        for proposal in proposals:
            data = await proposal.get_data()

            constraints = self._demand_offer_parser.parse_constraints(data.constraints)

            result.append(
                ProposalData(
                    properties=Properties(data.properties),
                    constraints=constraints,
                    proposal_id=data.proposal_id,
                    issuer_id=data.issuer_id,
                    state=data.state,
                    timestamp=cast(datetime, data.timestamp),
                    prev_proposal_id=data.prev_proposal_id,
                )
            )

        return result



from golem.managers.proposal.plugins.negotiating.add_payment_platform import (
    AddChosenPaymentPlatform,
)
from golem.managers.proposal.plugins.negotiating.negotiating_plugin import NegotiatingPlugin

__all__ = (
    "AddChosenPaymentPlatform",
    "NegotiatingPlugin",
)



import asyncio
import contextvars
import logging
from typing import Optional

from golem.utils.logging import trace_id_var

logger = logging.getLogger(__name__)


def create_task_with_logging(coro, *, trace_id: Optional[str] = None) -> asyncio.Task:
    context = contextvars.copy_context()
    task = context.run(_create_task_with_logging, coro, trace_id=trace_id)

    if trace_id is not None:
        task_name = trace_id
    else:
        task_name = task.get_name()

    logger.debug(f"Task `{task_name}` created")

    return task


def _create_task_with_logging(coro, *, trace_id: Optional[str] = None) -> asyncio.Task:
    if trace_id is not None:
        trace_id_var.set(trace_id)

    task = asyncio.create_task(coro)
    task.add_done_callback(_handle_task_logging)
    return task


def _handle_task_logging(task: asyncio.Task):
    try:
        return task.result()
    except asyncio.CancelledError:
        pass
    except Exception:
        logger.exception("Background async task encountered unhandled exception!")



import asyncio
import logging
from asyncio import Queue
from typing import List

from golem.managers import ProposalManagerPlugin
from golem.resources import Proposal
from golem.utils.asyncio import create_task_with_logging
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


class Buffer(ProposalManagerPlugin):
    def __init__(
        self,
        min_size: int,
        max_size: int,
        fill_concurrency_size: int = 1,
        fill_at_start=False,
    ):
        self._min_size = min_size
        self._max_size = max_size
        self._fill_concurrency_size = fill_concurrency_size
        self._fill_at_start = fill_at_start

        self._get_proposal_lock = asyncio.Lock()
        self._worker_tasks: List[asyncio.Task] = []
        # As there is no awaitable counter, queue with dummy values is used
        self._requests_queue: Queue[int] = asyncio.Queue()
        self._requests_pending_count = 0
        self._buffered: List[Proposal] = []
        self._buffered_condition = asyncio.Condition()
        self._is_started = False

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        if not self.is_started():
            raise RuntimeError("Not started!")

        async with self._get_proposal_lock:
            return await self._get_item()

    @trace_span()
    async def start(self) -> None:
        if self.is_started():
            raise RuntimeError("Already started!")

        for i in range(self._fill_concurrency_size):
            self._worker_tasks.append(
                create_task_with_logging(
                    self._worker_loop(), trace_id=get_trace_id_name(self, f"worker-{i}")
                )
            )

        if self._fill_at_start:
            self._handle_item_requests()

        self._is_started = True

    @trace_span()
    async def stop(self) -> None:
        if not self.is_started():
            raise RuntimeError("Already stopped!")

        for worker_task in self._worker_tasks:
            worker_task.cancel()

            try:
                await worker_task
            except asyncio.CancelledError:
                pass

        self._worker_tasks.clear()
        self._is_started = False

        self._requests_queue = asyncio.Queue()
        self._requests_pending_count = 0

    def is_started(self) -> bool:
        return self._is_started

    async def _worker_loop(self):
        while True:
            await self._wait_for_any_item_requests()

            item = await self._get_proposal()

            async with self._buffered_condition:
                self._buffered.append(item)

                self._buffered_condition.notify_all()

            self._requests_queue.task_done()
            self._requests_pending_count -= 1

    @trace_span()
    async def _wait_for_any_item_requests(self) -> None:
        await self._requests_queue.get()

    async def _get_item(self):
        async with self._buffered_condition:
            if self._get_items_count() == 0:  # This supports lazy (not at start) buffer filling
                logger.debug("No items to get, requesting fill")
                self._handle_item_requests()

            logger.debug("Waiting for any item to pick...")

            await self._buffered_condition.wait_for(lambda: 0 < len(self._buffered))
            item = self._buffered.pop()

            # Check if we need to request any additional items
            if self._get_items_count() < self._min_size:
                self._handle_item_requests()

            return item

    def _get_items_count(self):
        return len(self._buffered) + self._requests_pending_count

    @trace_span()
    def _handle_item_requests(self) -> None:
        items_to_request = self._max_size - self._get_items_count()

        for i in range(items_to_request):
            self._requests_queue.put_nowait(i)

        self._requests_pending_count += items_to_request

        logger.debug("Requested %d items", items_to_request)
